@@ -9,18 +9,17 @@ import base64
 import cgi
 import hashlib
 import hmac
-import imp
 import os
 import re
 import sys
 import time
+import threading
 import traceback
 from datetime import datetime, timedelta
-from email.utils import formatdate
+from email.utils import mktime_tz, parsedate_tz
 from functools import update_wrapper, wraps
 from mimetypes import guess_type as guess_ct
-from threading import Lock
-from wsgiref.util import FileWrapper
+from wsgiref.handlers import format_date_time, FileWrapper
 try:                  # Python 3
     from http.client import responses as HTTP_CODES
     from http.cookies import SimpleCookie
@@ -37,21 +36,22 @@ except ImportError:   # Python 2
     recode = lambda s: s.decode('utf-8')
 
 DEFAULT_BIND = {'host': '127.0.0.1', 'port': 8080}
-SECRET_KEY = None
 MAIN_MODULE = '__main__'
 
-__version__ = '0.2'
+__version__ = '0.3'
 __all__ = ['HTTPError', 'BadRequest', 'Forbidden', 'NotFound',  # HTTP errors
            'MethodNotAllowed', 'InternalServerError', 'Redirect',
            # Base classes
-           'HTTPHeaders', 'EnvironHeaders', 'Request', 'Response', 'Fiole',
+           'Accept', 'HTTPHeaders', 'EnvironHeaders', 'Request', 'Response',
            # Decorators
            'route', 'get', 'post', 'put', 'delete', 'errorhandler',
            # Template engine and static file helper
            'Loader', 'Lexer', 'Parser', 'BlockBuilder', 'Engine', 'Template',
            'engine', 'get_template', 'render_template', 'send_file',
            # WSGI application and server
-           'default_app', 'get_app', 'run_fiole']
+           'Fiole', 'default_app', 'get_app', 'run_wsgiref', 'run_fiole']
+_accept_re = re.compile(r'(?:^|,)\s*([^\s;,]+)(?:[^,]*?;\s*q=([\d.]*))?')
+_new_module = type(re)
 
 
 # Exceptions
@@ -105,9 +105,10 @@ def tobytes(value):
     return value.encode('utf-8') if isinstance(value, unicode) else value
 
 
-def html_escape(s):
+def escape_html(s):
     """Escape special chars in HTML string."""
-    return cgi.escape(s).replace('"', '&quot;').replace("'", '&#x27;')
+    return (cgi.escape(s if isinstance(s, unicode) else unicode(s))
+               .replace('"', '&quot;').replace("'", '&#x27;'))
 
 
 def format_timestamp(ts):
@@ -117,36 +118,18 @@ def format_timestamp(ts):
     elif isinstance(ts, (tuple, time.struct_time)):
         ts = time.mktime(ts)
     try:
-        return formatdate(ts, usegmt=True)
+        return format_date_time(ts)
     except Exception:
         raise TypeError("Unknown timestamp type: %r" % ts)
 
 
 def compare_digest(a, b):
     result = len(a) ^ len(b)
-    for x, y in zip(a, b):
+    for (x, y) in zip(a, b):
         result |= ord(x) ^ ord(y)
     return not result
 
 
-def _create_signed_value(secret, name, value):
-    value = base64.b64encode(tobytes(value)).decode('utf-8')
-    timestamp = '%X' % time.time()
-    signature = _create_signature(secret, name, value, timestamp)
-    return (value + '|' + timestamp + '|' + signature)
-
-
-def _decode_signed_value(secret, name, value, max_age_days=31):
-    parts = (value or '').split("|")
-    if len(parts) != 3:
-        return  # Invalid
-    if time.time() - int(parts[1], 16) > max_age_days * 86400:
-        return  # Expired
-    signature = _create_signature(secret, name, parts[0], parts[1])
-    if compare_digest(parts[2], signature):
-        return base64.b64decode(parts[0].encode('ascii')).decode('utf-8')
-
-
 def _create_signature(secret, *parts):
     sign = hmac.new(tobytes(secret), digestmod=hashlib.sha1)
     for part in parts:
@@ -206,6 +189,89 @@ def lock_acquire(f):
     return wrapper
 
 
+class Accept(object):
+    """Represent an ``Accept``-style header."""
+
+    def __init__(self, header_name, value):
+        accept_all = (not value and header_name != 'Accept-Encoding')
+        self._parsed = list(self.parse(value)) if value else []
+        self._masks = [('*', 1)] if accept_all else self._parsed
+        if header_name == 'Accept-Language':
+            self._match = self._match_language
+
+    @staticmethod
+    def parse(value):
+        """Parse ``Accept``-style header."""
+        for match in _accept_re.finditer(value):
+            (name, quality) = match.groups()
+            if name != 'q':
+                try:
+                    quality = float(quality or 1.)
+                    if quality:
+                        yield (name.lower(), min(1, quality))
+                except ValueError:
+                    yield (name.lower(), 1)
+
+    def __bool__(self):
+        return bool(self._parsed)
+    __nonzero__ = __bool__
+
+    def __iter__(self):
+        for (mask, q) in sorted(self._parsed, key=lambda m: -m[1]):
+            yield mask
+
+    def __contains__(self, offer):
+        """Return True if the given offer is listed in the accepted types."""
+        assert '*' not in offer
+        for (mask, q) in self._masks:
+            if self._match(mask, offer):
+                return True
+
+    def quality(self, offer):
+        """Return the quality of the given offer."""
+        assert '*' not in offer
+        bestq = 0
+        for (mask, q) in self._parsed:
+            if bestq < q and self._match(mask, offer):
+                bestq = q
+        return bestq or None
+
+    def best_match(self, offers, default_match=None):
+        """Return the best match in the sequence of offered types."""
+        best_offer = (default_match, -1, '*')
+        for offer in offers:
+            server_quality = 1
+            if isinstance(offer, (tuple, list)):
+                (offer, server_quality) = offer
+            assert '*' not in offer
+            for (mask, q) in self._masks:
+                possible_quality = server_quality * q
+                if (possible_quality > best_offer[1] or
+                    (possible_quality == best_offer[1] and
+                     mask.count('*') < best_offer[2].count('*')
+                     )) and self._match(mask, offer):
+                    best_offer = (offer, possible_quality, mask)
+        return best_offer[0]
+
+    @staticmethod
+    def _match(mask, offer):
+        if mask.endswith('/*'):
+            (mask, offer) = (mask[:-2], offer.split('/')[0])
+        return (mask == '*' or mask == offer.lower())
+
+    @staticmethod
+    def _match_language(mask, offer):
+        offer = offer.replace('_', '-').lower()
+        return (mask == '*' or mask == offer or
+                offer.startswith(mask + '-') or mask.startswith(offer + '-'))
+
+    @classmethod
+    def header(cls, header):
+        def fget(request):
+            return cls(header, request.headers[header])
+        return lazyproperty(fget)
+
+
 class HTTPHeaders(object):
     """An object that stores some headers."""
 
@@ -308,7 +374,7 @@ class EnvironHeaders(HTTPHeaders):
         return self.environ.get('HTTP_' + key)
 
     def __iter__(self):
-        for key, value in self.environ.items():
+        for (key, value) in self.environ.items():
             if key.startswith('HTTP_'):
                 if key not in ('HTTP_CONTENT_TYPE', 'HTTP_CONTENT_LENGTH'):
                     yield (key[5:].replace('_', '-').title(), value)
@@ -336,10 +402,18 @@ class Request(object):
             self.content_length = int(self.headers['Content-Length'] or 0)
         except ValueError:
             self.content_length = 0
+    accept = Accept.header('Accept')
+    accept_charset = Accept.header('Accept-Charset')
+    accept_encoding = Accept.header('Accept-Encoding')
+    accept_language = Accept.header('Accept-Language')
 
     def __getattr__(self, name):
         """Access the environment."""
-        return self.environ[name]
+        try:
+            return self.environ[name]
+        except KeyError:
+            raise AttributeError("type object %r has no attribute %r" %
+                                 (self.__class__.__name__, name))
 
     @lazyproperty
     def GET(self):
@@ -377,15 +451,15 @@ class Request(object):
         """Return the given signed cookie if it validates, or None."""
         if value is None:
             value = self.get_cookie(name)
-        return _decode_signed_value(SECRET_KEY, name, value,
-                                    max_age_days=max_age_days)
+        app = self.environ['fiole.app']
+        return app.decode_signed(name, value, max_age_days=max_age_days)
 
     def build_get_dict(self):
         """Take GET data and rip it apart into a dict."""
         raw_query_dict = parse_qs(self.query, keep_blank_values=1)
         query_dict = {}
 
-        for key, value in raw_query_dict.items():
+        for (key, value) in raw_query_dict.items():
             query_dict[key] = value if len(value) > 1 else value[0]
         return query_dict
 
@@ -413,66 +487,68 @@ class Response(object):
                  content_type='text/html', wrapped=False):
         self.output, self.status, self.wrapped = output, status, wrapped
         self.headers = HTTPHeaders(headers)
-        if ';' not in content_type and (
-                content_type.startswith('text/') or
-                content_type == 'application/xml' or
-                (content_type.startswith('application/') and
-                 content_type.endswith('+xml'))):
-            content_type += '; charset=' + self.charset
-        self.headers['Content-Type'] = content_type
+        if status != 304 and 'Content-Type' not in self.headers:
+            if ';' not in content_type and (
+                    content_type.startswith('text/') or
+                    content_type == 'application/xml' or
+                    (content_type.startswith('application/') and
+                     content_type.endswith('+xml'))):
+                content_type += '; charset=' + self.charset
+            self.headers['Content-Type'] = content_type
 
     def set_cookie(self, name, value, domain=None, expires=None, path="/",
-                   expires_days=None, **kwargs):
+                   expires_days=None, signed=None, **kwargs):
         """Set the given cookie name/value with the given options."""
         name = str(name)
         value = value if isinstance(value, str) else value.encode('utf-8')
         if re.search(r"[\x00-\x20]", name + value):
             raise ValueError("Invalid cookie %r: %r" % (name, value))
-        if not hasattr(self, "_new_cookie"):
-            self._new_cookie = SimpleCookie()
-        if name in self._new_cookie:
-            del self._new_cookie[name]
-        self._new_cookie[name] = value
-        morsel = self._new_cookie[name]
-        if domain:
-            morsel["domain"] = domain
         if expires_days is not None and not expires:
             expires = datetime.utcnow() + timedelta(days=expires_days)
-        if expires:
-            morsel["expires"] = format_timestamp(expires)
-        if path:
-            morsel["path"] = path
-        for (k, v) in kwargs.items():
-            if k == 'max_age':
-                k = 'max-age'
-            morsel[k] = v
+        attrs = [("domain", domain), ("path", path),
+                 ("expires", expires and format_timestamp(expires))]
+        if "max_age" in kwargs:
+            attrs.append(("max-age", kwargs.pop("max_age")))
+        if not hasattr(self, "_new_cookie"):
+            self._new_cookie = SimpleCookie()
+        elif name in self._new_cookie:
+            del self._new_cookie[name]
+        self._new_cookie[name] = value if not signed else ""
+        morsel = self._new_cookie[name]
+        for (key, val) in attrs + list(kwargs.items()):
+            morsel[key] = val or ""
+        morsel._signed = signed and value
 
     def clear_cookie(self, name, path="/", domain=None):
         """Delete the cookie with the given name."""
-        expires = datetime.utcnow() - timedelta(days=365)
-        self.set_cookie(name, value="", path=path, expires=expires,
-                        domain=domain)
+        self.set_cookie(
+            name, value="", path=path, expires_days=(-365), domain=domain)
 
     def set_secure_cookie(self, name, value, expires_days=30, **kwargs):
         """Sign and timestamp a cookie so it cannot be forged."""
-        self.set_cookie(name, self.create_signed_value(name, value),
-                        expires_days=expires_days, **kwargs)
-
-    def create_signed_value(self, name, value):
-        """Return a signed string with timestamp."""
-        return _create_signed_value(SECRET_KEY, name, value)
+        self.set_cookie(
+            name, value, expires_days=expires_days, signed=True, **kwargs)
 
     def send(self, environ, start_response):
         """Send the headers and return the body of the response."""
         status = "%d %s" % (self.status, HTTP_CODES.get(self.status))
-        output = self.output if self.wrapped else [tobytes(self.output or '')]
+        body = (self.output if self.wrapped else
+                [tobytes(self.output)]) if self.output else []
         if not self.wrapped:
-            self.headers['Content-Length'] = str(len(output[0]))
+            self.headers['Content-Length'] = str(body and len(body[0]) or 0)
         if hasattr(self, "_new_cookie"):
+            app = environ['fiole.app']
             for cookie in self._new_cookie.values():
+                if cookie._signed is not None:
+                    self._new_cookie[cookie.key] = (
+                        app.encode_signed(cookie.key, cookie._signed))
                 self.headers.add("Set-Cookie", cookie.OutputString(None))
         start_response(status, self.headers.to_list())
-        return output if (environ['REQUEST_METHOD'] != 'HEAD') else ()
+        if environ['REQUEST_METHOD'] != 'HEAD':
+            return body
+        if hasattr(body, 'close'):
+            body.close()
+        return []
 
 
 class Fiole(object):
@@ -482,7 +558,9 @@ class Fiole(object):
 
     def __init__(self):
         self.routes = []
+        self.hooks = []
         self.error_handlers = {302: http_302_found}
+        self.debug = False
 
     @classmethod
     def push(cls, app=None):
@@ -497,30 +575,40 @@ class Fiole(object):
 
     def handle_request(self, environ, start_response):
         """The main handler.  Dispatch to the user's code."""
+        environ['fiole.app'] = self
+        request = Request(environ)
+        hooks = [hdl(request) for hdl in self.hooks]
         try:
-            request = Request(environ)
-            (callback, kwargs, status) = self.find_matching_url(request)
-            response = callback(request, **kwargs)
-        except Exception as exc:
-            (response, status) = self.handle_error(exc, environ)
-        if not isinstance(response, Response):
-            response = Response(response, status=status)
-        return response.send(environ, start_response)
+            try:
+                for hook in hooks:
+                    hook.send(None)                 # Pre-process the Request
+                (callback, kwargs, status) = self.find_matching_url(request)
+                response = callback(request, **kwargs)
+            except Exception as exc:
+                (response, status) = self.handle_error(exc, environ)
+            if not isinstance(response, Response):
+                response = Response(response, status=status)
+            for hook in reversed(hooks):
+                response = hook.send(response)      # Post-process the Response
+            return response.send(environ, start_response)
+        finally:
+            for hook in reversed(hooks):
+                hook.close()                        # Clean up
 
     def handle_error(self, exception, environ, level=0):
         """Deal with the exception and present an error page."""
+        status = getattr(exception, 'status', 500)
+        if level > 2 or self.debug and status == 500:
+            raise
         if not getattr(exception, 'hide_traceback', False):
             environ['wsgi.errors'].write("%s occurred on '%s': %s\n%s" % (
                 exception.__class__.__name__, environ['PATH_INFO'],
                 exception, traceback.format_exc()))
-        status = getattr(exception, 'status', 500)
         handler = (self.error_handlers.get(status) or
                    self.default_error_handler(status))
         try:
             return handler(exception), status
-        except Exception as exc:
-            if level > 3:
-                raise
+        except HTTPError as exc:
             return self.handle_error(exc, environ, level + 1)
 
     def find_matching_url(self, request):
@@ -537,6 +625,24 @@ class Fiole(object):
                                    "not supported." % request.method)
         raise NotFound("Sorry, nothing here.")
 
+    def encode_signed(self, name, value):
+        """Return a signed string with timestamp."""
+        value = base64.b64encode(tobytes(value)).decode('utf-8')
+        timestamp = '%X' % time.time()
+        signature = _create_signature(self.secret_key, name, value, timestamp)
+        return (value + '|' + timestamp + '|' + signature)
+
+    def decode_signed(self, name, value, max_age_days=31):
+        """Decode a signed string with timestamp or return ``None``."""
+        try:
+            (value, timestamp, signed) = value.split('|')
+        except (ValueError, AttributeError):
+            return  # Invalid
+        signature = _create_signature(self.secret_key, name, value, timestamp)
+        if (compare_digest(signed, signature) and
+                time.time() - int(timestamp, 16) < max_age_days * 86400):
+            return base64.b64decode(value.encode('ascii')).decode('utf-8')
+
     def send_file(self, request, filename, root=None, content_type=None,
                   buffer_size=64 * 1024):
         """Fetch a static file from the filesystem."""
@@ -552,9 +658,15 @@ class Fiole(object):
         if not os.access(desired_path, os.R_OK):
             raise Forbidden("You do not have permission to access this file.")
         stat = os.stat(desired_path)
+        try:
+            ims = parsedate_tz(request.headers['If-Modified-Since'].strip())
+        except Exception:
+            ims = None
+        if ims and int(stat.st_mtime) <= mktime_tz(ims):    # 304 Not Modified
+            return Response(None, status=304, wrapped=True)
+
         headers = {'Content-Length': str(stat.st_size),
                    'Last-Modified': format_timestamp(stat.st_mtime)}
-
         if not content_type:
             content_type = guess_ct(filename)[0] or 'application/octet-stream'
         file_wrapper = request.environ.get('wsgi.file_wrapper', FileWrapper)
@@ -635,10 +747,10 @@ HTTP_CODES[511] = "Network Authentication Required"
 
 SPECIAL_TOKENS = 'extends require # include import from def end'.split()
 TOKENS = {'end' + k: 'end' for k in 'for if while with try class def'.split()}
-TOKENS.update({k: 'block' for k in 'for if while with try class'.split()})
+TOKENS.update({k: 'compound' for k in 'for if while with try class'.split()})
 TOKENS.update({k: 'continue' for k in 'else elif except finally'.split()})
 TOKENS.update({k: k for k in SPECIAL_TOKENS})
-COMPOUND_TOKENS = {'extends', 'def', 'block', 'continue'}
+COMPOUND_TOKENS = {'extends', 'def', 'compound'}
 OUT_TOKENS = {'markup', 'var', 'include'}
 isidentifier = re.compile(r'[a-zA-Z_]\w*').match
 setdefs = "super_defs['?'] = ?; ? = local_defs.setdefault('?', ?)".replace
@@ -662,12 +774,12 @@ class Loader(object):
     def load(self, name, source=None):
         """Return template by name."""
         if source is not None:
-            return unicode(source)
+            return (name, unicode(source))
         if name in self.templates:
-            return self.templates[name]
+            return (name, self.templates[name])
         path = os.path.join(_get_root_folder(), self.template_folder, name)
         with open(path, 'rb') as f:
-            return f.read().decode('utf-8')
+            return (path, f.read().decode('utf-8'))
 
 
 class Lexer(object):
@@ -743,15 +855,16 @@ class Parser(Lexer):
         """If token is ``continue`` prepend it with ``end`` token so
         it simulates a closed block.
         """
-        for lineno, token, value in tokens:
+        for (lineno, token, value) in tokens:
             if token == 'continue':
                 yield lineno, 'end', None
+                token = 'compound'
             yield lineno, token, value
 
     def parse_iter(self, tokens):
         """Process and yield groups of tokens."""
         operands = []
-        for lineno, token, value in tokens:
+        for (lineno, token, value) in tokens:
             if token in OUT_TOKENS:
                 operands.append((lineno, token, value))
                 continue
@@ -763,25 +876,25 @@ class Parser(Lexer):
                 if token != 'extends' and vals and vals[-1][1] != 'end':
                     raise SyntaxError('Missing "end" statement at line %d.' %
                                       vals[-1][0])
-                yield lineno, token, (value, vals)
-            else:
-                yield lineno, token, value
-                if token == 'end':
-                    break
+                value = (value, vals)
+            yield lineno, token, value
+            if token == 'end':
+                break
         if operands:
             yield operands[0][0], 'out', operands
 
 
 class BlockBuilder(list):
 
-    filters = {'s': 'str', 'e': 'escape'}
+    filters = {'e': 'escape'}
     writer_declare = '_b = []; w = _b.append'
-    writer_return = "return ''.join(_b)"
+    writer_return = 'return "".join(_b)'
 
-    def __init__(self, indent='', lineno=0, nodes=()):
+    def __init__(self, indent='', lineno=0, nodes=(), default_filters=None):
         self.indent = indent
         self.lineno = self.offset = lineno
         self.local_vars = set()
+        self.default_filters = default_filters or []
         self.build_block(nodes)
 
     def __enter__(self):
@@ -790,7 +903,7 @@ class BlockBuilder(list):
     def __exit__(self, exc_type, exc_value, exc_tb):
         self.indent = self.indent[:-4]
 
-    def add(self, lineno, code, token='statement'):
+    def add(self, lineno, code):
         """Add Python code to the source."""
         assert lineno >= self.lineno
         if code == 'return':
@@ -810,14 +923,14 @@ class BlockBuilder(list):
         return self     # Convenient for the context manager API
 
     def build_block(self, nodes):
-        for lineno, token, value in nodes:
+        for (lineno, token, value) in nodes:
             self.build_token(lineno, value, token)
         return True
 
     def build_token(self, lineno, value, token):
         assert token in self.rules, ('No rule to build "%s" token '
                                      'at line %d.' % (token, lineno))
-        return any(r(self, lineno, value, token) for r in self.rules[token])
+        return any(r(self, lineno, value) for r in self.rules[token])
 
     def compile_code(self, name):
         """Compile the generated source code."""
@@ -827,155 +940,128 @@ class BlockBuilder(list):
 
     # all builder rules
 
-    def build_extends(self, lineno, nodes, token):
-        assert token == 'render'
-        if len(nodes) not in (1, 2):    # Ignore 'require' before 'extends'
-            return
-        lineno, token, value = nodes[-1]
-        if token != 'extends':
-            return
-        extends, nodes = value
-        stmt = 'return _r(' + extends + ', ctx, local_defs, super_defs)'
-        self.build_block([n for n in nodes if n[1] in ('def', 'require')])
-        return self.add(self.lineno + 1, stmt)
-
-    def build_import(self, lineno, value, token):
-        assert token == 'import'
+    def build_import(self, lineno, value):
         parts = value[7:].rsplit(None, 2)
         if len(parts) == 3 and parts[1] == 'as':
             if parts[0] in self.local_vars or not isidentifier(parts[0]):
-                return self.add(lineno, parts[2] + ' = _i(' + parts[0] + ')')
+                value = "%s = _i(%s)" % (parts[2], parts[0])
         return self.add(lineno, value)
 
-    def build_from(self, lineno, value, token):
-        assert token == 'from'
-        name, tok2, var = value[5:].rsplit(None, 2)
+    def build_from(self, lineno, value):
+        (name, token, var) = value[5:].rsplit(None, 2)
         alias = var
-        if tok2 == 'as':
-            name, tok2, var = name.rsplit(None, 2)
-        assert tok2 == 'import'
+        if token == 'as':
+            (name, token, var) = name.rsplit(None, 2)
+        assert token == 'import'
         if name in self.local_vars or not isidentifier(name):
             value = "%s = _i(%s).local_defs['%s']" % (alias, name, var)
         return self.add(lineno, value)
 
-    def build_render_single_markup(self, lineno, nodes, token):
-        assert token == 'render' and lineno <= 0
+    def build_render(self, lineno, nodes):
+        assert lineno <= 0
         if not nodes:
             return self.add(lineno, "return ''")
+        # Ignore 'require' before 'extends'
+        if len(nodes) < 3 and nodes[-1][1] == 'extends':
+            (extends, nodes) = nodes[-1][2]
+            stmt = 'return _r(' + extends + ', ctx, local_defs, super_defs)'
+            self.build_block([n for n in nodes if n[1] in ('def', 'require')])
+            return self.add(self.lineno + 1, stmt)
         if len(nodes) == 1:
-            ln, token, nodes = nodes[0]
-            if token == 'out' and len(nodes) == 1:
-                ln, token, value = nodes[0]
+            (ln, token, subnodes) = nodes[0]
+            if token == 'out' and len(subnodes) == 1:
+                (ln, token, value) = subnodes[0]
                 if token == 'markup':
                     return self.add(ln, "return %r" % value)
-
-    def build_render(self, lineno, nodes, token):
-        assert token == 'render' and lineno <= 0
         self.add(lineno, self.writer_declare)
         self.build_block(nodes)
         return self.add(self.lineno + 1, self.writer_return)
 
-    def build_def_syntax_check(self, lineno, value, token):
-        assert token == 'def'
-        stmt, nodes = value
-        lineno, token, value = nodes[0]
+    def build_def(self, lineno, value):
+        (stmt, nodes) = value
+        (ln, token, subnodes) = nodes[0]
         if token in COMPOUND_TOKENS:
-            token = token.rstrip()
             error = ("The compound statement '%s' is not allowed here. "
                      "Add a line before it with %%#ignore.\n\n%s\n"
                      "    %%#ignore\n    %%%s ..." % (token, stmt, token))
             with self.add(lineno, stmt):
-                return self.add(lineno, 'raise SyntaxError(%r)' % error)
-
-    def build_def_single_markup(self, lineno, value, token):
-        assert token == 'def'
-        stmt, nodes = value
-        if len(nodes) > 2:
-            return
-        if len(nodes) == 2:
-            ln, token, nodes = nodes[0]
-            if token != 'out' or len(nodes) > 1:
-                return
-            ln, token, value = nodes[0]
-            if token != 'markup':
-                return
-            value = repr(value)
-        else:
-            ln, value = lineno, "''"
+                return self.add(ln, 'raise SyntaxError(%r)' % error)
+        single_markup = None
+        if len(nodes) == 1:
+            (ln, single_markup) = (lineno, '')
+        elif len(nodes) == 2:
+            if token == 'out' and len(subnodes) == 1:
+                (ln, token, value) = subnodes[0]
+                if token == 'markup':
+                    single_markup = value
         with self.add(lineno, stmt):
-            self.add(ln, "return " + value)
+            if single_markup is None:
+                self.add(lineno + 1, self.writer_declare)
+                self.build_block(nodes)
+                ln = self.lineno
+                self.add(ln, self.writer_return)
+            else:
+                self.add(ln, "return " + repr(single_markup))
         return self.add(ln + 1, setdefs('?', stmt[4:stmt.index('(', 5)]))
 
-    def build_def(self, lineno, value, token):
-        assert token == 'def'
-        stmt, nodes = value
-        with self.add(lineno, stmt):
-            self.add(lineno + 1, self.writer_declare)
-            self.build_block(nodes)
-            ln = self.lineno
-            self.add(ln, self.writer_return)
-        return self.add(ln + 1, setdefs('?', stmt[4:stmt.index('(', 5)]))
-
-    def build_out(self, lineno, nodes, token):
-        assert token == 'out'
-        for lineno, token, value in nodes:
+    def build_out(self, lineno, nodes):
+        for (lineno, token, value) in nodes:
             if token == 'include':
                 value = '_r(' + value + ', ctx, local_defs, super_defs)'
             elif token == 'var':
-                if '|' in value:
-                    filters = [f.strip() for f in value.split('|')]
-                    value = filters.pop(0)
-                    for f in filters:
-                        value = self.filters.get(f, f) + '(' + value + ')'
+                filters = [f.strip() for f in value.split('|')]
+                value = filters.pop(0)
+                for f in filters + self.default_filters:
+                    value = self.filters.get(f, f) + '(' + value + ')'
             elif value:
                 value = repr(value)
             if value:
                 self.add(lineno, 'w(' + value + ')')
 
-    def build_compound(self, lineno, value, token):
-        assert token in COMPOUND_TOKENS
-        stmt, nodes = value
+    def build_compound(self, lineno, value):
+        (stmt, nodes) = value
         with self.add(lineno, stmt):
             return self.build_block(nodes)
 
-    def build_require(self, lineno, values, token):
-        stmt = '; '.join([v + " = ctx['" + v + "']"
+    def build_require(self, lineno, values):
+        stmt = "; ".join([v + " = ctx['" + v + "']"
                           for v in values if v not in self.local_vars])
         self.local_vars.update(values)
         return self.add(lineno, stmt)
 
-    def build_end(self, lineno, value, token):
+    def build_end(self, lineno, value):
         if self.lineno != lineno:
             self.add(lineno - 1, '')
 
-    rules = {
-        'render': [build_extends, build_render_single_markup, build_render],
-        'import': [build_import],
-        'from': [build_from],
-        'require': [build_require],
-        'out': [build_out],
-        'def': [build_def_syntax_check, build_def_single_markup, build_def],
-        'block': [build_compound],
-        'continue': [build_compound],
-        'end': [build_end],
-        'statement': [add],
-        '#': [],
-    }
+    rules = {'statement': [add], '#': []}
+    for name in 'import from render require out compound def end'.split():
+        rules[name] = [locals()['build_' + name]]
 
 
 class Engine(object):
     """Assemble the template engine."""
 
     def __init__(self, loader=None, parser=None, template_class=None):
-        self.lock = Lock()
-        self.templates, self.renders, self.modules = {}, {}, {}
-        self.global_vars = {'_r': self.render, '_i': self.import_name}
+        self.lock = threading.Lock()
+        self.clear()
+        self.default_filters = ['str']
+        self.global_vars = {'_r': self.render, '_i': self.import_name,
+                            'str': unicode, 'escape': escape_html}
         self.template_class = template_class or Template
         self.loader = loader or Loader()
         self.parser = parser or Parser()
+        self.build = BlockBuilder
+
+    def clear(self):
+        """Remove all compiled templates from the internal cache."""
+        self.templates, self.renders, self.modules = {}, {}, {}
 
     def get_template(self, name=None, **kwargs):
-        """Return a compiled template."""
+        """Return a compiled template.
+
+        The optional keyword argument ``default_filters`` overrides
+        the setting which is configured on the ``Engine``.
+        """
         if name and kwargs:
             self.remove(name)
         try:
@@ -985,7 +1071,7 @@ class Engine(object):
 
     @lock_acquire
     def remove(self, name):
-        """Remove given ``name`` from internal cache."""
+        """Remove given ``name`` from the internal cache."""
         if name in self.renders:
             del self.templates[name], self.renders[name]
         if name in self.modules:
@@ -1011,11 +1097,11 @@ class Engine(object):
     def compile_template(self, name, **kwargs):
         if name in self.templates:
             return self.templates[name]
-        nodes = self.load_and_parse(name, **kwargs)
+        (path, nodes, filters) = self.load_and_parse(name, **kwargs)
         def_render = 'def render(ctx, local_defs, super_defs):'
-        nodes = [(-1, 'block', (def_render, [(0, 'render', list(nodes))]))]
-        source = BlockBuilder(lineno=-2, nodes=nodes)
-        compiled = source.compile_code(name or '<string>')
+        nodes = [(-1, 'compound', (def_render, [(0, 'render', list(nodes))]))]
+        source = self.build(lineno=-2, nodes=nodes, default_filters=filters)
+        compiled = source.compile_code(path or '<string>')
         local_vars = {}
         exec(compiled, self.global_vars, local_vars)
         template = self.template_class(name, local_vars['render'])
@@ -1026,20 +1112,23 @@ class Engine(object):
 
     @lock_acquire
     def compile_import(self, name, **kwargs):
-        if name not in self.modules:
-            nodes = self.load_and_parse(name, **kwargs)
-            nodes = ([(-1, 'statement', 'local_defs = {}; super_defs = {}')] +
-                     [n for n in nodes if n[1] == 'def'])
-            source = BlockBuilder(lineno=-2, nodes=nodes)
-            compiled = source.compile_code(name)
-            self.modules[name] = module = imp.new_module(name)
-            module.__dict__.update(self.global_vars)
-            exec(compiled, module.__dict__)
+        if name in self.modules:
+            return
+        (path, nodes, filters) = self.load_and_parse(name, **kwargs)
+        nodes = ([(-1, 'statement', 'local_defs = {}; super_defs = {}')] +
+                 [n for n in nodes if n[1] == 'def'])
+        source = self.build(lineno=-2, nodes=nodes, default_filters=filters)
+        compiled = source.compile_code(path)
+        self.modules[name] = module = _new_module(name)
+        module.__dict__.update(self.global_vars)
+        exec(compiled, module.__dict__)
 
     def load_and_parse(self, name, **kwargs):
-        template_source = self.loader.load(name, **kwargs)
+        filters = kwargs.pop('default_filters', self.default_filters)
+        (path, template_source) = self.loader.load(name, **kwargs)
         tokens = self.parser.tokenize(template_source)
-        return self.parser.parse_iter(self.parser.end_continue(tokens))
+        nodes = self.parser.parse_iter(self.parser.end_continue(tokens))
+        return (path, nodes, filters)
 
 
 class Template(object):
@@ -1055,11 +1144,12 @@ class Template(object):
         return self.render_template(ctx or kwargs, {}, {})
 
 engine = Engine()
-engine.global_vars.update({'str': unicode, 'escape': html_escape})
 
 
 def get_template(name=None, source=None, require=None):
     """Return a compiled template."""
+    if get_app().debug:
+        engine.clear()
     if source is None:
         if '\n' not in name and '{{' not in name:
             return engine.get_template(name)
@@ -1077,24 +1167,33 @@ def render_template(template_name=None, source=None, **context):
 # The WSGI HTTP server
 
 def run_wsgiref(host, port, handler):
-    from wsgiref.simple_server import make_server
+    """Simple HTTPServer that supports WSGI."""
+    from wsgiref.simple_server import make_server, ServerHandler
+
+    def cleanup_headers(self):
+        # work around http://bugs.python.org/issue18099
+        if self.status[:3] == "304":
+            del self.headers['Content-Length']
+        elif 'Content-Length' not in self.headers:
+            self.set_content_length()
+    ServerHandler.cleanup_headers = cleanup_headers
+
     srv = make_server(host, port, handler)
     srv.serve_forever()
 
 
-def run_fiole(app=None, server=run_wsgiref, host=None, port=None, secret=None):
+def run_fiole(app=default_app, server=run_wsgiref, host=None, port=None):
     """Run the *Fiole* web server."""
-    global SECRET_KEY
-    SECRET_KEY = secret or base64.b64encode(os.urandom(32))
-
-    assert (app or default_app).routes, "No route defined"
+    if not hasattr(app, 'secret_key'):
+        app.secret_key = base64.b64encode(os.urandom(33))
+    assert app.routes, "No route defined"
     host = host or DEFAULT_BIND['host']
     port = int(port or DEFAULT_BIND['port'])
     print('`fiole` starting up (using %s)...\nListening on http://%s:%s...\n'
-          'Use Ctrl-C to quit.\n' % (server, host, port))
+          'Use Ctrl-C to quit.\n' % (server.__name__, host, port))
 
     try:
-        server(host, port, (app or default_app).handle_request)
+        server(host, port, app.handle_request)
     except KeyboardInterrupt:
         print('\nShutting down.  Have a nice day!')
         raise SystemExit(0)
